@@ -96,3 +96,11 @@ class RoleCreationFailure(Exception):
 
 class NoActiveArtifactRetrieved(Exception):
     """Could not retrieve an active artifact."""
+
+
+class NonExistentOU(Exception):
+    """The OU name provided does not exist in Control Tower."""
+
+
+class InvalidParentHierarchy(Exception):
+    """The parent hierarchy provided is not valid."""
